/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include <string.h>
#include "mex.h"
#include "frame.h"
#include "stream.h"
#include "synth.h"
#include "layer3.h"

#define INPUT_BUFFER_SIZE	(5*8192)
#define MP3_GRANULARITY 4600

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {

	/* Declarations */
	char *filename;
	int filename_len;
	FILE *mp3;
	struct mad_stream Stream;
	struct mad_frame Frame;
	struct mad_synth Synth;
	unsigned char InputBuffer[INPUT_BUFFER_SIZE+MAD_BUFFER_GUARD];
	int i,j;
	double *data = NULL;
	int data_size;
	int numframes,numsamples,delay;
	int numchannels = 2;
	unsigned int samplerate = 0;
	unsigned long bitrate = 0;
	double scale = pow(2,-28);

	/* Check number of arguments */
	if (nrhs<1) mexErrMsgTxt( "Not enough input arguments.");
	if (nlhs>3) mexErrMsgTxt( "Too many output arguments.");

	/* Input */
	filename_len = mxGetN(prhs[0])+1;
	filename = (char*)mxCalloc(filename_len, sizeof(char));
	mxGetString(prhs[0],filename,filename_len);
	mp3 = fopen(filename,"r");
	if (mp3==NULL) mexErrMsgTxt( "Cannot open file\n");

	/* Delay */
	delay = 2257;
	if (nrhs>1) delay = (int)mxGetPr(prhs[1])[0];
	if (delay<0) mexErrMsgTxt( "Invalid delay\n");

	/* Decoding */
	mad_stream_init(&Stream);
	mad_frame_init(&Frame);
	mad_synth_init(&Synth);
	numframes=0;
	if (nlhs>0) {
		data_size = 1152*MP3_GRANULARITY;
		data = (double*) mxCalloc(data_size,sizeof(double));
	}

	do {
		fail:
		if(Stream.buffer==NULL || Stream.error==MAD_ERROR_BUFLEN) {
			size_t ReadSize,Remaining;
			unsigned char *ReadStart;
			if(Stream.next_frame!=NULL) {
				Remaining=Stream.bufend-Stream.next_frame;
				memmove(InputBuffer,Stream.next_frame,Remaining);
				ReadStart=InputBuffer+Remaining;
				ReadSize=INPUT_BUFFER_SIZE-Remaining;
			} else { 
				ReadSize=INPUT_BUFFER_SIZE,
				ReadStart=InputBuffer,
				Remaining=0;
			}
			ReadSize=fread(ReadStart,1,ReadSize,mp3);
			if(ReadSize<=0) break;
			mad_stream_buffer(&Stream,InputBuffer,ReadSize+Remaining);
			Stream.error = MAD_ERROR_NONE;
		}
		Frame.options = Stream.options;
		if (!(Frame.header.flags & MAD_FLAG_INCOMPLETE) && mad_header_decode(&Frame.header, &Stream) == -1) goto fail;

		if (numframes==0) {
			if (Frame.header.layer!=3) mexErrMsgTxt( "File is not MPEG-1 Layer 3.\n");
			bitrate = Frame.header.bitrate;
			samplerate = Frame.header.samplerate;
			if (Frame.header.mode==0) numchannels = 1;
		}

		Frame.header.flags &= ~MAD_FLAG_INCOMPLETE;
		mad_layer_III(&Stream, &Frame, 0);
		mad_synth_frame(&Synth, &Frame);

		if (nlhs>0) {
			if ((numframes*1152+1152)*numchannels>=data_size) {
				data_size += 1152*MP3_GRANULARITY;
				data = (double*) mxRealloc(data, data_size*sizeof(double));
			}
			for (i=0 ; i<1152 ; i++) {
				for (j=0;j<numchannels;j++)
					data[(numframes*1152+i)*numchannels+j] = (double)Synth.pcm.samples[j][i];
			}
		}

		numframes++;
	}while(1);

	/* End decoding */
	mad_frame_finish(&Frame);
	mad_stream_finish(&Stream);
	mad_synth_finish(&Synth);

	/* Output */
	if (nlhs>0) {
		numsamples = numframes*1152 - delay;
		plhs[0] = mxCreateDoubleMatrix(numsamples,numchannels,mxREAL);
		for (i=0;i<numsamples;i++)
			for (j=0;j<numchannels;j++)
				mxGetPr(plhs[0])[j*numsamples+i] = data[i*numchannels+j+delay*numchannels]*scale;
	}
	if (nlhs>1) {
		plhs[1] = mxCreateDoubleMatrix(1,1,mxREAL);
		mxGetPr(plhs[1])[0] = (unsigned int)samplerate;
	}
	if (nlhs>2) {
		plhs[2] = mxCreateDoubleMatrix(1,1,mxREAL);
		mxGetPr(plhs[2])[0] = (unsigned long)bitrate;
	}

	/* Clean */
	if (data) mxFree( data );
	fclose( mp3 );
	mxFree( filename );
}
